! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: region_aux
!
! Authors:  Masaaki Sakano (SSC/LUX)
!   Originally transferred from region_mod.f90, developed by
!	    Richard West - SSC (Leicester)
!	    Ian Stewart (SSC/LUX)
!          
! Description: Core subroutines/functions in region.
! 
! Comments: 
! 
! $Id: region_aux.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module region_aux

  use dal	! For type definition of single etc.
  use cal
  ! use param
  use errorHandling
  ! use datafilter

  use array_utils	! for smallest/largestTrueIndex
  use confusion	! for findConfusedSets()
  use ellipse	! for EllipseT

  implicit none 

  private 

  public getBandEdges
  public initializeAnnulusType
  public getEncircEnergyRadii
  public correctRadii
  public setOverlapFractions
  public calcSepAndOverlapFrac
  public correctConfSet
  public findMinPosMaxOverlap
  public findMaxNegOverlap
  public shrinkRadiusEffective
  public adjustOverlaps

  public localAnnulusType

  type :: localAnnulusType
    real(double) ::&
      centreX,&
      centreY
    real(single) ::&
      innerLongAxis,&
      outerLongAxis
    type(EllipseT) ::& ! typedef in ssclib/ellipse
      ellipse
    character(10) :: unit	! eg 'pixels', 'detxy', 'tanxy'
    logical(bool) :: isAnnulus	! FALSE means the inner area is of zero size.
    ! This should override any non-zero value in innerLongAxis.
  end type localAnnulusType

contains 

  !---------------------------------------------------------------------
  subroutine getBandEdges(set, bandLo, bandHi, infoFoundThisBand)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(in) :: set
    real(double), intent(out) :: bandLo(3,0:9), bandHi(3,0:9)
    logical,      intent(out) :: infoFoundThisBand(3,0:9)
    ! logical(bool), intent(out) :: infoFoundThisBand(3,0:9)

    character(2) :: instChars(3)
    integer :: idInst, idBand
    character(8) :: loKwdName, hiKwdName

    !-------------------------------------------------------------------
    ! Start

    instChars = (/'PN', 'M1', 'M2'/)

    infoFoundThisBand = .false. ! default
    bandLo = 0.0 ! default
    bandHi = 0.0 ! default
    do idInst = 1, 3
      do idBand = 1, 9
        write(loKwdName, '(a3, i1, a4)') instChars(idInst)//'_', idBand, '_ELO'
        write(hiKwdName, '(a3, i1, a4)') instChars(idInst)//'_', idBand, '_EHI'
        if (hasAttribute(set, loKwdName)&
        .and. hasAttribute(set, hiKwdName)) then
          infoFoundThisBand(idInst,idBand) = .true.
          bandLo(idInst,idBand) = real64Attribute(set, loKwdName)
          bandHi(idInst,idBand) = real64Attribute(set, hiKwdName)
        end if
      end do

      if (any(infoFoundThisBand(idInst,1:9))) then 
        infoFoundThisBand(idInst,0) = .true.
        bandLo(idInst,0) = bandLo(idInst&
          , smallestTrueIndex(infoFoundThisBand(idInst,1:9)))
        bandHi(idInst,0) = bandHi(idInst&
          ,  largestTrueIndex(infoFoundThisBand(idInst,1:9)))
        ! These calls are in ssclib/array_utils.
      end if
    end do

  end subroutine getBandEdges



  !---------------------------------------------------------------------
  subroutine initializeAnnulusType(annulus)
  !---------------------------------------------------------------------

    implicit none

    type(localAnnulusType), intent(out) :: annulus(:)

    !-------------------------------------------------------------------
    ! Start

    annulus%centreX = 0.0d0
    annulus%centreY = 0.0d0
    annulus%innerLongAxis = 0.0
    annulus%outerLongAxis = 0.0
    call initializeEllipseT(annulus%ellipse) ! in ssclib/ellipse
    annulus%unit = 'none'
    annulus%isAnnulus = .true. ! false means that the inner ellipse is of
    ! zero size.

  end subroutine initializeAnnulusType


  !---------------------------------------------------------------------
  subroutine getEncircEnergyRadii(srcThetaArcsec, srcPhi, srcEeV&
  , fracEncircEnergy, srcRadiusTan)
  !---------------------------------------------------------------------
  ! NOTE! Assumes (i) that cal has been properly set up for the instrument.

    implicit none

    real(double), intent(in)  :: srcPhi(:),&
                                 srcThetaArcsec(size(srcPhi)),&
                                 srcEeV(size(srcPhi))
    real(single), intent(in)  :: fracEncircEnergy
    real(single), intent(out) :: srcRadiusTan(size(srcPhi))

    real(single) ::&
      srcRadiusMm(size(srcPhi))
    integer(int16) ::&
      numSources, i
    type(PsfT) ::&
      psf
    real(double) ::&
      focalLengthMm
    type(PsfValidityRangesT) :: validPsfRanges
    real(double) :: effectiveSrcThetaArcsec

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcPhi)

    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)
    validPsfRanges = CAL_psfValidityRanges()
!    if (validPsfRanges%thetaMax > 900.0) then
!      write(errStr, '(a, es12.5, a)') "The maximum off-axis angle (theta) &
!        &returned from the CCF is given as", validPsfRanges%thetaMax&
!        , " arcsec. It ought to be 900. You should use more recent CCF &
!        &components. If you persist with this set, the task will fail with &
!        &an error 'OffAxisAngleOutOfValidBounds' for any source which has &
!        &an off-axis angle greater than 900 arcsec."
!      call warning('badCcfConstituent')
!    end if
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLengthMm)

    do i = 1, numSources
      if  (srcEeV(i) < validPsfRanges%eMin&
      .or. srcEeV(i) > validPsfRanges%eMax&
      .or. srcThetaArcsec(i) < validPsfRanges%thetaMin&
      .or. srcPhi(i) < validPsfRanges%phiMin&
      .or. srcPhi(i) > validPsfRanges%phiMax)&
call error('energyThetaOrPhiOutOfRange',"One or more of energy, theta or &
&phi were outside allowed ranges for the PSF.")

      if (srcThetaArcsec(i) > validPsfRanges%thetaMax) then
        write(errStr, '(a, i5, a, es12.5, a)') "The off-axis angle (theta) for &
          &source ", i, " was outside the allowed range for the PSF ("&
          ,validPsfRanges%thetaMax," arcsec). The maximum valid value will &
          &be substituted."
        call warning('srcThetaTooLarge')

        effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
          * (1.0d0 - epsilon(srcThetaArcsec(i)))
      else
        effectiveSrcThetaArcsec = srcThetaArcsec(i)
      end if

      psf = CAL_getPSF(srcEeV(i), effectiveSrcThetaArcsec, srcPhi(i))
      srcRadiusMm(i) = CAL_psfEncircledEnergy(psf&
        , real(fracEncircEnergy, double))
    end do

    srcRadiusTan = srcRadiusMm / focalLengthMm

  end subroutine getEncircEnergyRadii



  !---------------------------------------------------------------------
  subroutine correctRadii(srcX, srcY, srcRadius, srcShrinkage)
  !---------------------------------------------------------------------
  ! Description:
  !  Returns srcShrinkage, where
  !  0 < srcShrinkage <= 1; the ratio of the returned size of the region
  !  to the original region (=srcData); that is, each region will be shrinked
  !  by the ratio of srcShrinkage, when there is a confusion with
  !  nearby sources.
  !
  ! Algorithm:
  !  1) Find confused "set" of sources.
  !  2) For each confused "set", examine the "subset" of the source data
  !    and calculate the ratio of shrinkage.
  !  3) Feedback to the original set from the subset, and loop to the Step(2).

    implicit none

    real(single),  dimension(:), intent(in) ::&
      srcX, srcY
    real(single),  dimension(:), intent(in) ::&
      srcRadius
    real(single),  dimension(:), intent(out) ::&
      srcShrinkage

    integer(int16) ::&
      numSources, i, j, confSetNum(size(srcX)), maxNumConfSets, numConfSrcs
    real(single),  dimension(size(srcX)) ::&
      tempSrcX, tempSrcY, tempSrcRadius, tempSrcRadiusShrinkage
logical, parameter :: localTest = .false.
logical :: test

    !-------------------------------------------------------------------
    ! Start

    srcShrinkage = 1.0 ! vector op, default (= no shrinkage).
  ! if (.not. adjustRadius) return	! adjustRadius is a command-line option; so it is now handled in the main routine.

    numSources = size(srcX)

    ! 1) Find confused "set" of sources.
    call findConfusedSets(real(srcX, double), real(srcY, double), real(srcRadius, double), confSetNum)
    ! in ssclib/confusion
    ! (See the comment-line of its source code for detail.
    !  In short, it returns the array "confSetNum", which contains
    !  either 0 (=no confusion) or positive indices for each element.)

    maxNumConfSets = maxval(confSetNum)	! Number of set of confused sources.
  if (maxNumConfSets .le. 0) return	! There is no confusion at all.

    do i = 1, maxNumConfSets
    if (count(confSetNum .eq. i) .le. 0) cycle	! Just to play safe.  This should not happen by the specification of findConfusedSets() in ssclib/confusion.

      ! 2a) For each confused "set", create the "subset" of the source data to examine.
      !   This is mainly for the sake of speed of processing,
      !   but partly also so that it would not change or set the srcShrinkage
      !   of the sources, which are not confused with the set (="i"-th set)
      !   of sources it is examining now.
      numConfSrcs = 0
      do j = 1, numSources
        if (confSetNum(j) .eq. i .and. srcRadius(j) .gt. 0.0) then
          numConfSrcs = numConfSrcs + 1
          tempSrcX(numConfSrcs)               = srcX(j)
          tempSrcY(numConfSrcs)               = srcY(j)
          tempSrcRadius(numConfSrcs)          = srcRadius(j)
          tempSrcRadiusShrinkage(numConfSrcs) = srcShrinkage(j)	! nb. srcShrinkage(j) may be set already, I mean, may be different from the initial value of 1.0.
        end if
      end do

      ! 2b) For each subset, examine the "subset" of the source data
      !    and calculate the ratio of shrinkage.
if (localTest) print *,'=== before correctConfSet'
test = .false.
      call correctConfSet(tempSrcX(1:numConfSrcs), tempSrcY(1:numConfSrcs)&
        , tempSrcRadius(1:numConfSrcs)&
        , tempSrcRadiusShrinkage(1:numConfSrcs),test)
      ! NOTE: Although tempSrcRadius() may contain some irrelevant data,
      ! which may have been set in the previous loop,
      ! it would not affect the processing here,
      ! because the indices given to correctConfSet() are
      ! limited to (1:numConfSrcs).
if (localTest) print *,'=== after correctConfSet'

      ! 3) Feedback to the original set from the subset, and loop.
      !  NOTE: This follows the same way of looping as in Step(2a).
      numConfSrcs = 0
      do j = 1, numSources
        if (confSetNum(j) .eq. i .and. srcRadius(j) .gt. 0.0) then
          numConfSrcs = numConfSrcs + 1
          srcShrinkage(j) = tempSrcRadiusShrinkage(numConfSrcs)
        end if
      end do
    end do

  end subroutine correctRadii


  !---------------------------------------------------------------------
  subroutine setOverlapFractions(srcRadiusEffective &
     , separations, overlapFraction, indexI, indexJ)
  !---------------------------------------------------------------------
  ! Core routine to (re-)calculate overlapFraction(i,j)
  ! from the separations between sources and their effective radii.
  ! 
  ! This routine sets (or modifies) the input
  !   overlapFraction(indexI, indexJ) and 
  !   overlapFraction(indexJ, indexI),
  ! but untouches the rest of overlapFraction(:,:).

    implicit none

    real(single), intent(in) :: srcRadiusEffective(:)
    real(single), intent(in) ::&
      separations(size(srcRadiusEffective), size(srcRadiusEffective))
    real(single), intent(inout) ::&
      overlapFraction(size(srcRadiusEffective), size(srcRadiusEffective))
    integer,      intent(in) :: indexI, indexJ
    real(single) :: sumOfRadii, overlap

!logical, parameter :: localTest = .true.
logical, parameter :: localTest = .false.

    !-------------------------------------------------------------------
    ! Start

    if (srcRadiusEffective(indexI) == 0) then
      write(errStr, '(a, i5, a)') "srcRadiusEffective(" &
         , indexI, ") is zero.  Contact the code developer."
      call error('zeroSrcRadiusEffective')
    end if
    if (srcRadiusEffective(indexJ) == 0) then
      write(errStr, '(a, i5, a)') "srcRadiusEffective(" &
         , indexJ, ") is zero.  Contact the code developer."
      call error('zeroSrcRadiusEffective')
    end if

    sumOfRadii = srcRadiusEffective(indexI) + srcRadiusEffective(indexJ)
    overlap = sumOfRadii - separations(indexI, indexJ)
       
    if (abs(overlap) &
        > epsilon( max(abs(srcRadiusEffective(indexI)), abs(srcRadiusEffective(indexJ))) ) * 100) then
      ! In order to ensure that the value of overlap is significant. 
      ! Here it compares with 100 times of epsilon, the almost negligible value, of the values of srcRadiusEffective
      !
      ! NOTE: This value of 100 (times) seems a bit too large.
      !  The reason of the choice is "make test" would fail in the SSC/LUX
      ! with a much smaller value, such as 10, whereas it succeeds
      ! in the other SAS-group platforms even without this condition altogether.
      !  Whether it is 1 or 100, it is anyway a very small number
      ! compared with the parent values, so I am now leaving it.
      !  However it had better be investigated further.

      overlapFraction(indexI, indexJ) = overlap / srcRadiusEffective(indexI)
      overlapFraction(indexJ, indexI) = overlap / srcRadiusEffective(indexJ)
    else
      overlapFraction(indexI, indexJ) = 0.0
      overlapFraction(indexJ, indexI) = 0.0
    end if

    ! overlapFraction: (i,j)<0 means that source i and j are not confused;
    ! 0<(i,j)<1 means that i and j are confused but centre of i is
    ! outside radius of j (not necessarily the other way around);
    ! (i,j)>1 means that i and j are confused and centre of i is
    ! inside the radius of j (but not necessarily the other way around).

  end subroutine setOverlapFractions


  !---------------------------------------------------------------------
  subroutine calcSepAndOverlapFrac(srcX, srcY, srcRadiusEffective &
     , separations, overlapFraction)
  !---------------------------------------------------------------------
  ! Calculate separations(:,:) and overlapFraction(:,:),
  ! from the source positions and their effective radii.

    implicit none

    real(single), dimension(:), intent(in) :: srcX, srcY
    real(single), dimension(size(srcX)), intent(in) :: srcRadiusEffective
    real(single), dimension(size(srcX), size(srcX)), intent(out) ::&
      separations, overlapFraction
    integer :: numSources, i, j
    ! real(single) :: sumOfRadii, dX, dY, overlap
    real(single) :: dX, dY

!logical, parameter :: localTest = .true.
logical, parameter :: localTest = .false.

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcX)

    ! Do some preliminary setup:
    separations = 0.0	! The distance between Src(i) and Src(j), therefore, it remains at this value (=0) on the diagonal, and always separation(i,j)==separation(j,i).
    overlapFraction = 0.0	! = (SumOfRadii-separations)/EffectiveRadius(i or j); therefore, it remains at this value (=0) on the diagonal.  See the comment below for more detail.
    do i = 1, numSources - 1
      do j = i + 1, numSources
        dX = srcX(i) - srcX(j)
        dY = srcY(i) - srcY(j)
        separations(i, j) = sqrt(dX * dX + dY * dY)
        separations(j, i) = separations(i, j)
        if (separations(i, j) .le. 0.0) then
          write(errStr, '(a, i5, a, i5, a, es12.5, a, es12.5, a)') 'Sources '&
          , i,' and ', j,' have the same position, namely (x,y) = ('&
          , srcX(i),',', srcY(i),'). You might want to filter the source &
          &list to remove sources with duplicate rows.'
call error('doubledSource')
        end if

        ! Set overlapFraction(i, j) and overlapFraction(j, i)
        call setOverlapFractions(srcRadiusEffective &
           , separations, overlapFraction, i, j)

        ! sumOfRadii = srcRadiusEffective(i) + srcRadiusEffective(j)
        ! overlap = sumOfRadii - separations(i, j)
        ! overlapFraction(i, j) = overlap / srcRadiusEffective(i)
        ! overlapFraction(j, i) = overlap / srcRadiusEffective(j)

        ! overlapFraction: (i,j)<0 means that source i and j are not confused;
        ! 0<(i,j)<1 means that i and j are confused but centre of i is
        ! outside radius of j (not necessarily the other way around);
        ! (i,j)>1 means that i and j are confused and centre of i is
        ! inside the radius of j (but not necessarily the other way around).
      end do
    end do

    if (localTest) then
      print *,'The values set by subroutine calcSepAndOverlapFrac():'
      print *,'srcRadiusEffective=',srcRadiusEffective
      print *,'separations=',separations
      print *,'overlapFraction=',overlapFraction
    end if

  end subroutine calcSepAndOverlapFrac


  !---------------------------------------------------------------------
  subroutine correctConfSet(srcX, srcY, srcRadius, srcShrinkage, test)
  !---------------------------------------------------------------------
  ! Description:
  !  For the given set of confused sources, this modifies srcShrinkage, where
  !  0 < srcShrinkage <= 1; the ratio of the returned size of the region
  !  to the original region; that is, each region will be shrinked
  !  by the ratio of srcShrinkage, when there is a confusion with
  !  nearby sources.
  !
  ! NOTE:
  !  Input srcShrinkage should have been already set: 0 < srcShrinkage <= 1.
  !  Input srcRadius is the original radius (or perhaps
  !  longer-semi-axis of ellipse; nb. the short-semi-axis
  !  is not taken into account in this routine).

    implicit none

    real(single),  dimension(:), intent(in) ::&
      srcX,&
      srcY,&
      srcRadius
    real(single),  dimension(:), intent(inout) ::&
      srcShrinkage
logical, intent(in), optional :: test

    real(single),  dimension(size(srcX), size(srcX)) ::&
      separations,&
      overlapFraction
    real(single),  dimension(size(srcX)) ::&
      srcRadiusEffective
    integer(int16) ::&
      numSources,&
      ! i,j&
      k,&
      iOfMinMax,&
      jOfMinMax
    real(single) ::&
      ! sumOfRadii,&
      ! dX,&
      ! dY,&
      ! overlap,&
      reduction, inflation
    logical :: noNegativeOverlaps, noPositiveOverlaps, localTest

    !-------------------------------------------------------------------
    ! Start

if (present(test)) then
  localTest = test
else
  localTest = .false.
end if

    srcRadiusEffective = srcRadius * srcShrinkage	! vector op.
if (minval(srcRadiusEffective) <= 0.0) call error('nonRealSrcRadii'&
, "The effective or post-shrinkage source radius is <= 0.")

    numSources = size(srcX)

    call calcSepAndOverlapFrac(srcX, srcY, srcRadiusEffective, separations &
     , overlapFraction)

!     ! Do some preliminary setup:
!     separations = 0.0	! The distance between Src(i) and Src(j), therefore, it remains at this value (=0) on the diagonal, and always separation(i,j)==separation(j,i).
!     overlapFraction = 0.0	! = (SumOfRadii-separations)/EffectiveRadius(i or j); therefore, it remains at this value (=0) on the diagonal.  See the comment below for more detail.
!     do i = 1, numSources - 1
!       do j = i + 1, numSources
!         dX = srcX(i) - srcX(j)
!         dY = srcY(i) - srcY(j)
!         separations(i, j) = sqrt(dX * dX + dY * dY)
!         separations(j, i) = separations(i, j)
!         if (separations(i, j) .le. 0.0) then
!           write(errStr, '(a, i5, a, i5, a, es12.5, a, es12.5, a)') 'Sources '&
!           , i,' and ', j,' have the same position, namely (x,y) = ('&
!           , srcX(i),',', srcY(i),'). You might want to filter the source &
!           &list to remove sources with duplicate rows.'
! call error('doubledSource')
!         end if
! 
!         sumOfRadii = srcRadiusEffective(i) + srcRadiusEffective(j)
!         overlap = sumOfRadii - separations(i, j)
!         overlapFraction(i, j) = overlap / srcRadiusEffective(i)
!         overlapFraction(j, i) = overlap / srcRadiusEffective(j)
! 
!         ! overlapFraction: (i,j)<0 means that source i and j are not confused;
!         ! 0<(i,j)<1 means that i and j are confused but centre of i is
!         ! outside radius of j (not necessarily the other way around);
!         ! (i,j)>1 means that i and j are confused and centre of i is
!         ! inside the radius of j (but not necessarily the other way around).
!       end do
!     end do

if (localTest) print *,'=== Before findMinPosMaxOverlap'
    ! Now selectively shrink sources in an attempt to remove all the overlaps:
    do while(maxval(overlapFraction) .gt. 0.0)
if (numSources < 6) then
else	! Specific debugging statment!!!  So DELETE THIS after debugging!!!!
!if (localTest) print *,'Bef:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac=',overlapFraction(iOfMinMax, jOfMinMax)
if (localTest) print *,'Bef:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac(2,6)=',overlapFraction(2,6)
if (localTest) print *,'Bef:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac(3,6)=',overlapFraction(3,6)
end if
      call findMinPosMaxOverlap(overlapFraction, noPositiveOverlaps&
      , iOfMinMax, jOfMinMax)
if (localTest) print *,'=== After findMinPosMaxOverlap: ', noPositiveOverlaps, maxval(overlapFraction)

    if (noPositiveOverlaps) exit

      ! Eliminate the overlap between sources iOfMinMax and jOfMinMax by
      ! shrinking them by an equal multiplicative factor (reduction):

      call shrinkRadiusEffective(srcRadiusEffective, separations&
         , iOfMinMax, jOfMinMax, localTest, overlapFraction, numSources, reduction)

!       reduction = separations(iOfMinMax, jOfMinMax)&
!       / (srcRadiusEffective(iOfMinMax) + srcRadiusEffective(jOfMinMax))
! 
!       reduction = min(max(reduction, 0.0), 1.0) ! just to make sure there are
!       ! no funny values.
! 
! if (numSources < 6) then
! else	! Specific debugging statment!!!  So DELETE THIS after debugging!!!!
! print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac(2,6)=',overlapFraction(2,6)
! print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac(3,6)=',overlapFraction(3,6)
! !print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac=',overlapFraction(iOfMinMax, jOfMinMax)
! end if
!       ! Modify srcRadiusEffective and accordingly recalculate overlapFraction.
!       srcRadiusEffective(iOfMinMax) = srcRadiusEffective(iOfMinMax) * reduction
!       srcRadiusEffective(jOfMinMax) = srcRadiusEffective(jOfMinMax) * reduction

!!!!!!!!!!!
! This adjustOverlaps() apparently reverses overlapFraction() and as a result,
!   overlapFraction(2,6)
!   overlapFraction(3,6)
! become zero, ALTERNATIVELY!!!
!!!!!!!!!!!
      call adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
        , iOfMinMax)
      call adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
        , jOfMinMax)

if (numSources < 3) then
      write(errStr, '(a, i5, a, f12.5, a, f12.5, a, es12.5, a, es12.5, a)') 'Source No.'&
         , iOfMinMax, '(x,y)=('&
         , srcX(iOfMinMax), ', ',srcY(iOfMinMax),') is shrinked in radius by (' &
         , reduction, '), now, or overall by (' &
         , srcRadiusEffective(iOfMinMax)/srcRadius(iOfMinMax), ').'
      call message(AppMsg, Noisy)
else	! Specific debugging statment!!!  So DELETE THIS after debugging!!!!
if (localTest) &
print *,'iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac=',overlapFraction(iOfMinMax, jOfMinMax)
      k=2
      write(errStr, '(a, i5, a, f12.5, a, f12.5, a, es12.5, a, es12.5, a)') 'Source No.'&
         , k, '(x,y)=('&
         , srcX(k), ', ',srcY(k),') is shrinked in radius by (' &
         , reduction, '), now, or overall by (' &
         , srcRadiusEffective(k)/srcRadius(k), ').'
      call message(AppMsg, Noisy)
      k=3
      write(errStr, '(a, i5, a, f12.5, a, f12.5, a, es12.5, a, es12.5, a)') 'Source No.'&
         , k, '(x,y)=('&
         , srcX(k), ', ',srcY(k),') is shrinked in radius by (' &
         , reduction, '), now, or overall by (' &
         , srcRadiusEffective(k)/srcRadius(k), ').'
      call message(AppMsg, Noisy)
end if
      write(errStr, '(a, i5, a, f12.5, a, f12.5, a, es12.5, a, es12.5, a)') 'Source No.'&
         , jOfMinMax, '(x,y)=('&
         , srcX(jOfMinMax), ', ',srcY(jOfMinMax),') is shrinked in radius by (' &
         , reduction, '), now, or overall by (' &
         , srcRadiusEffective(jOfMinMax)/srcRadius(jOfMinMax), ').'
      call message(AppMsg, Noisy)

      overlapFraction(iOfMinMax, jOfMinMax) = 0.0
      overlapFraction(jOfMinMax, iOfMinMax) = 0.0
if (numSources < 3) then
else	! Specific debugging statment!!!  So DELETE THIS after debugging!!!!
if (localTest) &
print *,'iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac=',overlapFraction(iOfMinMax, jOfMinMax)
end if

      ! Sources other than the pair just 'shrunk' can get 'left behind',
      ! that is, they can end up with radii that are unecessarily small.
      ! To prevent this, go through the overlapFractions and see if there is
      ! any with negative overlaps (indicating they are no longer confused);
      ! if so, choose that which has been shrunk the most and deshrink it.
      ! Repeat the exercise until there are no more sources which can be
      ! deshrunk.

!localTest=.true.
      do
        call findMaxNegOverlap(overlapFraction, srcRadiusEffective/srcRadius&
        , noNegativeOverlaps, iOfMinMax, jOfMinMax)

if (localTest) print *, 'iOfMinMax, jOfMinMax, noNegativeOverlaps:', iOfMinMax, jOfMinMax, noNegativeOverlaps
! if (localTest) print *, 'overlapFraction(iOfMinMax, jOfMinMax):', overlapFraction(iOfMinMax, jOfMinMax)	! iOfMinMax may be out of range.

      if (noNegativeOverlaps) exit

        inflation = (separations(iOfMinMax, jOfMinMax)&
        - srcRadiusEffective(jOfMinMax)) / srcRadiusEffective(iOfMinMax)

        inflation = max(inflation, 1.0) ! just to make sure there are
        ! no funny values.
if (localTest) print *, 'inflation:', inflation

        srcRadiusEffective(iOfMinMax) = min(srcRadius(iOfMinMax)&
        , srcRadiusEffective(iOfMinMax) * inflation)
if (localTest) print *, 'srcRadius(iOfMinMax):', srcRadius(iOfMinMax)
if (localTest) print *, 'srcRadiusEffective(iOfMinMax) * inflation:', srcRadiusEffective(iOfMinMax) * inflation
if (localTest) print *, 'srcRadiusEffective(iOfMinMax):', srcRadiusEffective(iOfMinMax)

        call adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
        , iOfMinMax)
      end do ! deshrinking loop.
    end do

    srcShrinkage = srcRadiusEffective / srcRadius	! vector op.

  end subroutine correctConfSet


  !---------------------------------------------------------------------
  subroutine findMinPosMaxOverlap(overlapFraction, noPositiveOverlaps&
  , iOfMinMax, jOfMinMax)
  !---------------------------------------------------------------------
  ! Description:
  !   Each source(i of (i,j)) has a overlapFraction to all the other sources.
  !  Here this routine returns the source number of iOfMinMax, whose source
  !  has the minimum overlapFraction value even for its maximum (on the
  !  condition that the maximum value is positive), and the corresponding
  !  jOfMinMax, which gives the maximum value for its iOfMinMax.
  !   If there is not a single positive overlapFraction, then this returns
  !  .true. for noPositiveOverlaps.
  ! 
  ! Example:
  !    j =   1    2    3       MAX
  !   i=1:  0.0  0.9  3.0      3.0
  !   i=2:  2.3  0.0  0.4  ==> 2.3
  !   i=3:  0.8  1.7  0.0      1.7  =>  iOfMinMax=3, jOfMinMax=2
  ! 

    implicit none

    real(single),   intent(in)  :: overlapFraction(:,:)
    integer(int16), intent(out) :: iOfMinMax, jOfMinMax
    logical, intent(out) :: noPositiveOverlaps

    integer(int16) :: i, numSources, locVector(1)
    real(single) ::&
      maxOverlapVector(size(overlapFraction, 1)),&
      minMaxOverlapFraction,&
tempVector(size(overlapFraction, 1))

    !-------------------------------------------------------------------
    ! Start

    numSources = size(overlapFraction, 1)

    noPositiveOverlaps = .true. ! default

    ! See if there are any sources with a positive maximum overlap: 
    do i = 1, numSources

tempVector = overlapFraction(i, :)
maxOverlapVector(i) = maxval(tempVector)

!      maxOverlapVector(i) = maxval(overlapFraction(i, 1:numSources))

      ! This now contains the maximum amount of overlap suffered by each
      ! source, except that this value will be zero for sources which actually
      ! have a negative maximum overlap (ie are not confused with any other),
      ! because overlapFraction always has zero values on the diagonal.
      ! However this does not matter here since we are only in this subroutine
      ! interested in the positive values.
    end do

    if (any(maxOverlapVector .gt. 0.0)) then
      noPositiveOverlaps = .false.

      ! Find the minimum positive value of maxOverlapVector:

      minMaxOverlapFraction = maxval(maxOverlapVector) ! starting value.
locVector = maxloc(maxOverlapVector)
iOfMinMax = locVector(1)
      do i = 1, numSources
        if (maxOverlapVector(i) .gt. 0.0&
        .and. maxOverlapVector(i) .lt. minMaxOverlapFraction) then
          iOfMinMax = i
          minMaxOverlapFraction = maxOverlapVector(i)
        end if
      end do

      ! Have now identified the source (iOfMinMax) with the lowest positive
      ! value of maxOverlapVector; now find which of the other sources
      ! overlaps iOfMinMax by the maximum amount:

tempVector = overlapFraction(iOfMinMax, :)
locVector = maxloc(tempVector)
!      locVector = maxloc(overlapFraction(iOfMinMax, :))
      jOfMinMax = locVector(1)
    end if

  end subroutine findMinPosMaxOverlap

  !---------------------------------------------------------------------
  subroutine findMaxNegOverlap(overlapFraction, srcShrinkage&
  , noNegativeOverlaps, iOfMinMax, jOfMinMax)
  !---------------------------------------------------------------------

    implicit none

    real(single),   intent(in)  :: overlapFraction(:,:), srcShrinkage(:)
    integer(int16), intent(out) :: iOfMinMax, jOfMinMax
    logical, intent(out) :: noNegativeOverlaps

    integer(int16) :: i, numSources,&
      locVector(1)
    real(single) ::&
      tempVector(size(srcShrinkage)),&
      maxOverlapVector(size(srcShrinkage)),&
      shrinkage(size(srcShrinkage))
real(single) :: myPrecision

    !-------------------------------------------------------------------
    ! Start
myPrecision = 1e-6
!******

    ! Initialisation (If (noNegativeOverlaps) then these values are returned.)
    iOfMinMax = -99
    jOfMinMax = -99

    numSources = size(srcShrinkage)

    noNegativeOverlaps = .true. ! default
    ! See if there are any sources with shrinkage and also a negative
    ! maximum overlap: 
    do i = 1, numSources
      tempVector = overlapFraction(i, :)
      tempVector(i) = minval(tempVector) ! to get rid of the meaningless
      ! zeroes on the diagonal of overlapFraction.

      maxOverlapVector(i) = maxval(tempVector)
      ! This now contains the maximum amount of overlap suffered by each
      ! source, including those sources with negative maximum overlap.
    end do

!    if (any(maxOverlapVector .lt. 0.0 .and. srcShrinkage .lt. 1.0)) then
if (any(maxOverlapVector .lt. -myPrecision .and. srcShrinkage .lt. 1.0 - myPrecision)) then
      noNegativeOverlaps = .false.

      ! For those sources which have a negative value of maxOverlapVector
      ! (ie are not now confused with any other), find that one which has
      ! suffered the maximum shrinkage:

      shrinkage = 1.0
!      where(maxOverlapVector .lt. 0.0)&
where(maxOverlapVector .lt. -myPrecision)&
        shrinkage = srcShrinkage

      locVector = minloc(shrinkage)
      iOfMinMax = locVector(1)

      ! Now for this source (iOfMinMax), find which other source is nearest
      ! to it. The gap between their radii will constrain by how much source
      ! iOfMinMax can be reinflated.

      tempVector = overlapFraction(iOfMinMax, :)
      tempVector(iOfMinMax) = minval(tempVector) ! to get rid of the meaningless
      ! zeroes on the diagonal of overlapFraction.
      locVector = maxloc(tempVector)
      jOfMinMax = locVector(1)
    end if

  end subroutine findMaxNegOverlap


  !---------------------------------------------------------------------
  subroutine shrinkRadiusEffective(srcRadiusEffective, separations&
  , iOfMinMax, jOfMinMax, localTestIn, overlapFraction, testNumSources, reductionOut)
  !---------------------------------------------------------------------
  ! Eliminate the overlap between sources iOfMinMax and jOfMinMax by
  ! shrinking them by an equal multiplicative factor (reduction):

    implicit none

    real(single),   intent(inout) :: srcRadiusEffective(:)
    real(single),   intent(in)    :: separations(:,:)
    integer(int16), intent(in)    :: iOfMinMax, jOfMinMax
    logical,        intent(in), optional :: localTestIn
    real(single),   intent(in), optional :: overlapFraction(:,:)	! For debugging, needed in this subroutine
    integer(int16), intent(in), optional :: testNumSources	! For debugging, needed in this subroutine
    real(single),   intent(out), optional :: reductionOut	! For debugging, maybe needed by the caller

    real(single) :: reduction, prevISrcRadiusEff, prevJSrcRadiusEff
    integer(int16) :: numSources
logical :: localTest

    !-------------------------------------------------------------------
    ! Start

    if (present(localTestIn)) then
      localTest = localTestIn
    else
      localTest = .false.
    end if

    if (present(testNumSources) .and. present(overlapFraction)) then
      numSources = testNumSources
      if (verbosity(AppMsg) >= 7) localTest = .true.
    else
      numSources = -1
    end if

      ! Eliminate the overlap between sources iOfMinMax and jOfMinMax by
      ! shrinking them by an equal multiplicative factor (reduction):

      reduction = separations(iOfMinMax, jOfMinMax)&
      / (srcRadiusEffective(iOfMinMax) + srcRadiusEffective(jOfMinMax))

!print *,'before-reduction'
      if (reduction >= 1.0) then
        write(errStr, '(a, i5, a, i5, a, es12.5)') &
             "No shrinkage happens for (i,j)=(" &
           , iOfMinMax, ",", jOfMinMax, "): reduction=", reduction
        call warning('noEffectiveReduction')
      else if (reduction <= 0.0) then
        write(errStr, '(a, i5, a, i5, a, es12.5)') &
             "Effective radius reduced to be zero for (i,j)=(" &
           , iOfMinMax, ",", jOfMinMax, "): reduction=", reduction
        call warning('noEffectiveReduction')
      end if

      reduction = min(max(reduction, 0.0), 1.0) ! just to make sure there are
      ! no funny values.

      if (present(reductionOut)) reductionOut = reduction

if (localTest) then	! For debug.
if (numSources < 6) then
else if (present(overlapFraction)) then	! Specific debugging statment!!!  So DELETE THIS after debugging!!!!
print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac(2,6)=',overlapFraction(2,6)
print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac(3,6)=',overlapFraction(3,6)
!print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac=',overlapFraction(iOfMinMax, jOfMinMax)
else
print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax
print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax
end if
end if

      prevISrcRadiusEff = srcRadiusEffective(iOfMinMax)	! For debug
      prevJSrcRadiusEff = srcRadiusEffective(jOfMinMax)	! For debug

      ! Modify srcRadiusEffective and accordingly recalculate overlapFraction.
      srcRadiusEffective(iOfMinMax) = srcRadiusEffective(iOfMinMax) * reduction
      srcRadiusEffective(jOfMinMax) = srcRadiusEffective(jOfMinMax) * reduction
if (localTest) then	! For debug.
if (numSources < 6) then
else  	! Specific debugging statment!!!  So DELETE THIS after debugging!!!!
print *,'I:srcRadiusEffective(,',iOfMinMax,')=', prevISrcRadiusEff, ' => ',srcRadiusEffective(iOfMinMax)
print *,'J:srcRadiusEffective(,',jOfMinMax,')=', prevJSrcRadiusEff, ' => ',srcRadiusEffective(jOfMinMax)
!print *,'Aft:iOfMinMax=',iOfMinMax,' jOfMinMax=',jOfMinMax,' overlapFrac=',overlapFraction(iOfMinMax, jOfMinMax)
end if
end if

  end subroutine shrinkRadiusEffective


  !---------------------------------------------------------------------
  subroutine adjustOverlaps(srcRadiusEffective, separations, overlapFraction&
  , iOfMinMax)
  !---------------------------------------------------------------------
  ! When srcRadiusEffective(iOne) changes, overlapFraction(:) related to
  ! the source iOne should be recalculated.  That is what this routine does.

    implicit none

    real(single),   intent(in)    :: srcRadiusEffective(:), separations(:,:)
    real(single),   intent(inout) :: overlapFraction(:,:)
    integer(int16), intent(in)    :: iOfMinMax

    integer(int16) :: i, numSources
    ! real(single)   :: sumOfRadii, overlap

!logical, parameter :: localTest = .true.
logical, parameter :: localTest = .false.

    !-------------------------------------------------------------------
    ! Start

    numSources = size(srcRadiusEffective)

    if (localTest) then
      print *,'iOfMinMax =', iOfMinMax
      print *,'srcRadiusEffective =', srcRadiusEffective(:)
      do i = 1, numSources
        print *, 'separations(',i,',:)=', separations(i,:)
      end do
      do i = 1, numSources
        print *, 'overlapFraction(',i,',:)=', overlapFraction(i,:)
      end do
    end if

    do i = 1, numSources
    if (i .eq. iOfMinMax) cycle
      ! Set overlapFraction(i, iOfMinMax) and overlapFraction(iOfMinMax, i)
      call setOverlapFractions(srcRadiusEffective &
         , separations, overlapFraction, int(i), int(iOfMinMax))

      ! sumOfRadii = srcRadiusEffective(i) + srcRadiusEffective(iOfMinMax)
      ! overlap = sumOfRadii - separations(i, iOfMinMax)
      ! overlapFraction(i, iOfMinMax) = overlap / srcRadiusEffective(i)
      ! overlapFraction(iOfMinMax, i) = overlap / srcRadiusEffective(iOfMinMax)
    end do

  end subroutine adjustOverlaps

end module region_aux
