package TestRegion;
# ESA (C) 2000-2018 
#
# This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
#
#    SAS is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    SAS is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

our $VERSION = '1.1.0';

use strict;
use Data::Dumper;
use SSCLib;
use Fitsplutils;

=head1 NAME

TestRegion - Common subroutines for testing region.

=head1 SYNOPSIS

    use TestRegion

=head1 DESCRIPTION

Every subroutine is called as

  my $ret = &TestRegion::THIS_SUBROUTINE(...);

=item prepare_data ( )

=item getisrun ("01100", [Running-Processes-In-Default])

=cut
  ;	# For Emacs


############################################################


## Global variables
our (%evtfits, %orgsrclist, %outsrcregfits, %outbkgregfits);
our $Defaultdir = "testdata";
our $Defaultoutdir = "outfiles";

%evtfits =
  ( 'basic_m1' => "$Defaultdir/P0022140101M1S005MIEVLI0000.FTZ",
    'basic_m2' => "$Defaultdir/P0022140101M2S006MIEVLI0000.FTZ",
    'basic_pn' => "$Defaultdir/P0022140101PNS004PIEVLI0000.FTZ" );
%orgsrclist =
  ( 'basic_m1' => "$Defaultdir/EPICSourceProducts-emos1-1-Filtered_ML_src_list-X00000000000000.ftz",
    'basic_m2' => "$Defaultdir/EPICSourceProducts-emos2-1-Filtered_ML_src_list-X00000000000000.ftz",
    'basic_pn' => "$Defaultdir/EPICSourceProducts-epn-1-Filtered_ML_src_list-X00000000000000.ftz");

%outsrcregfits =
 ( 'basic_m1' => "$Defaultoutdir/pps_m1_srcreg.ds",
   'basic_m2' => "$Defaultoutdir/pps_m2_srcreg.ds",
   'basic_pn' => "$Defaultoutdir/pps_pn_srcreg.ds",
   'batch_m1' => "$Defaultoutdir/batch_m1_srcreg.ds",
   'batch_m2' => "$Defaultoutdir/batch_m2_srcreg.ds",
   'batch_pn' => "$Defaultoutdir/batch_pn_srcreg.ds");	# Output
%outbkgregfits =
 ( 'basic_m1' => "$Defaultoutdir/pps_m1_bkgreg.ds",
   'basic_m2' => "$Defaultoutdir/pps_m2_bkgreg.ds",
   'basic_pn' => "$Defaultoutdir/pps_pn_bkgreg.ds",
   'batch_m1' => "$Defaultoutdir/batch_m1_bkgreg.ds",
   'batch_m2' => "$Defaultoutdir/batch_m2_bkgreg.ds",
   'batch_pn' => "$Defaultoutdir/batch_pn_bkgreg.ds" );	# Output

#---------------------------------------------------------------------
sub prepare_data {
#---------------------------------------------------------------------
  ## Input: NONE
  ## Return: 0 if normal ends.
  ## Description: mainly gunzip the data, if needed.

  my ($i, $vname);
  my (@files);
  my ($stat) = 0;
  # my (@var2chk) = ( \%evtfits, \%orgsrclist, \%outsrcregfits, \%outbkgregfits );
  my (@var2chk) = ( \%evtfits, \%orgsrclist );
  my $data_dir = &SSCLib::gettestdata_dir("ssclib","with_src_prods");

  foreach $vname (@var2chk) {	# Loop over (global) variable names.
    foreach (keys(%$vname)) {	#  Loop over each key in each variable.
      push(@files, $$vname{$_});	# Gets each filename to process.
    }
  }

  # foreach (keys(%evtfits)) {
  #   push(@files, $evtfits{$_});
  # }
  # foreach (keys(%orgorgsrclist)) {
  #   push(@files, $orgorgsrclist{$_});
  # }
  # foreach (keys(%outsrcregfits)) {
  #   push(@files, $outsrcregfits{$_});
  # }
  # foreach (keys(%outbkgregfits)) {
  #   push(@files, $outbkgregfits{$_});
  # }

  my $statnow = 0;
  foreach (@files) {
    $statnow = abs( &SSCLib::gunzipfits_if_needed($_, $data_dir) );
    print STDERR "Error in preparing ($_)\n" if $statnow;
    $stat += $statnow;
  }

  return($stat);
}


# #---------------------------------------------------------------------
# sub initarray {
# #---------------------------------------------------------------------
#   ## Input: (Size or Array)
#   ## Return: ($UNDEFINEDNUMBER, $UNDEFINEDNUMBER, ...)
#   ## Description: If Array is given, return the initialised (number)
#   ##   array with the same size as the argument.
#   ## Global var(main): $UNDEFINEDNUMBER
# 
#   my ($size) = scalar(@_);
#   if ( $size <= 1) {
#     $size = shift;
#   }
#   my (@arret);
# 
#   foreach (0..($size-1)) {
#     push(@arret, $main::UNDEFINEDNUMBER);
#   }
# 
#   return @arret;
# }
# 

#---------------------------------------------------------------------
sub getisrun {
#---------------------------------------------------------------------
  ## Input: ($runbin, Default)	# Binary of "11011" etc, where the highest is for the first.
  ## Return: (@isrun)	# Array of run(1) or not(0)
  ## Description: Return the array which shows 1 or 0 for each
  ##   index.  The first one ($isrun[0]) is always "".

  my($runstr) = shift; 
  my($runstrdef) = shift; 
  my(@isrun);

  $runstrdef = "11111111111111111111111111111" if ((! defined($runstrdef)) || ($runstrdef eq ""));
  $runstr = $runstrdef if ((! defined($runstr)) || ($runstr eq ""));
  my($a) = length($runstrdef)-length($runstr);
  if ($a > 0) {
    $runstr .= substr($runstrdef, length($runstr), $a);	# Fill with Defaults where not specified 
  }

  @isrun = split('', $runstr);
  unshift @isrun, "";

  return @isrun;	# ("", 1, 0, 1, 1, 1) etc.

  ## Unit-Test: &TestRegion::getisrun("11", "10111") => ( '', '1', '1', '1', '1', '1' )
  ## Unit-Test: &TestRegion::getisrun("11", "10101") => ( '', '1', '1', '1', '0', '1' )
  ## Unit-Test: &TestRegion::getisrun("001","10111") => ( '', '0', '0', '1', '1', '1' )
}


1;
